#include "owb.h"
//---------------------------------------------------------------------
static const char *TAG = "owb";
//---------------------------------------------------------------------
#define OW_DEBUG
//---------------------------------------------------------------------
#define OW_DURATION_RESET 480
#define OW_DURATION_SLOT 75
#define OW_DURATION_RX_IDLE (OW_DURATION_SLOT + 2)
//--------------------------------------------------------------------
#define info_of_driver(owb) container_of(owb, owb_rmt_driver_info, bus)
//---------------------------------------------------------------------
static bool _is_init(const OneWireBus * bus)
{
  bool ok = false;
  if (bus != NULL)
  {
    if (bus->driver)
    {
      // OK
      ok = true;
    }
    else
    {
      ESP_LOGE(TAG, "bus is not initialised");
    }
  }
  else
  {
    ESP_LOGE(TAG, "bus is NULL");
  }
  return ok;
}
//---------------------------------------------------------------------
owb_status owb_use_crc(OneWireBus * bus, bool use_crc)
{
    owb_status status = OWB_STATUS_NOT_SET;

    if (!bus)
    {
        status = OWB_STATUS_PARAMETER_NULL;
    }
    else if (!_is_init(bus))
    {
        status = OWB_STATUS_NOT_INITIALIZED;
    }
    else
    {
        bus->use_crc = use_crc;
        ESP_LOGD(TAG, "use_crc %d", bus->use_crc);

        status = OWB_STATUS_OK;
    }

    return status;
}
//---------------------------------------------------------------------
static void onewire_flush_rmt_rx_buf(const OneWireBus * bus)
{
    void * p = NULL;
    size_t s = 0;

    owb_rmt_driver_info * i = info_of_driver(bus);

    while ((p = xRingbufferReceive(i->rb, &s, 0)))
    {
        ESP_LOGD(TAG, "flushing entry");
        vRingbufferReturnItem(i->rb, p);
    }
}
//---------------------------------------------------------------------
static owb_status _search(const OneWireBus * bus, OneWireBus_SearchState * state, bool * is_found)
{
  bool search_result = false;
  owb_status status = OWB_STATUS_NOT_SET;

  // if the last call was not the last one
  if (!state->last_device_flag)
  {
    // 1-Wire reset
    bool is_present;
    bus->driver->reset(bus, &is_present);

    if (!is_present)
    {
        // reset the search
        state->last_discrepancy = 0;
        state->last_device_flag = false;
        state->last_family_discrepancy = 0;
        *is_found = false;
        return OWB_STATUS_OK;
    }
  }




  status = OWB_STATUS_OK;

  *is_found = search_result;

  return status;
}
//---------------------------------------------------------------------
owb_status owb_search_first(const OneWireBus * bus, OneWireBus_SearchState * state, bool * found_device)
{
    bool result;
    owb_status status = OWB_STATUS_NOT_SET;

    if (!bus || !state || !found_device)
    {
        status = OWB_STATUS_PARAMETER_NULL;
    }
    else if (!_is_init(bus))
    {
        status = OWB_STATUS_NOT_INITIALIZED;
    }
    else
    {
        memset(&state->rom_code, 0, sizeof(state->rom_code));
        state->last_discrepancy = 0;
        state->last_family_discrepancy = 0;
        state->last_device_flag = false;
        _search(bus, state, &result);
        status = OWB_STATUS_OK;

        *found_device = result;
    }

    return status;
}
//---------------------------------------------------------------------
static owb_status _uninitialize(const OneWireBus *bus)
{
  owb_rmt_driver_info * info = info_of_driver(bus);

  rmt_driver_uninstall(info->tx_channel);
  rmt_driver_uninstall(info->rx_channel);

  return OWB_STATUS_OK;
}
//---------------------------------------------------------------------
static owb_status _reset(const OneWireBus * bus, bool * is_present)
{
  rmt_item32_t tx_items[1] = {0};
  bool _is_present = false;
  int res = OWB_STATUS_OK;
  owb_rmt_driver_info * i = info_of_driver(bus);

  tx_items[0].duration0 = OW_DURATION_RESET;
  tx_items[0].level0 = 0;
  tx_items[0].duration1 = 0;
  tx_items[0].level1 = 1;

  uint16_t old_rx_thresh = 0;
  rmt_get_rx_idle_thresh(i->rx_channel, &old_rx_thresh);
  rmt_set_rx_idle_thresh(i->rx_channel, OW_DURATION_RESET + 60);

  onewire_flush_rmt_rx_buf(bus);

  rmt_rx_start(i->rx_channel, true);

  if (rmt_write_items(i->tx_channel, tx_items, 1, true) == ESP_OK)
  {
    size_t rx_size = 0;
    rmt_item32_t * rx_items = (rmt_item32_t *)xRingbufferReceive(i->rb, &rx_size, 100 / portTICK_PERIOD_MS);

    if (rx_items)
    {
      if (rx_size >= (1 * sizeof(rmt_item32_t)))
      {
#ifdef OW_DEBUG
        ESP_LOGI(TAG, "rx_size: %d", rx_size);

        for (int i = 0; i < (rx_size / sizeof(rmt_item32_t)); i++)
        {
          ESP_LOGI(TAG, "i: %d, level0: %d, duration %d", i, rx_items[i].level0, rx_items[i].duration0);
          ESP_LOGI(TAG, "i: %d, level1: %d, duration %d", i, rx_items[i].level1, rx_items[i].duration1);
        }
#endif
        // parse signal and search for presence pulse
        if ((rx_items[0].level0 == 0) && (rx_items[0].duration0 >= OW_DURATION_RESET - 2))
        {
          if ((rx_items[0].level1 == 1) && (rx_items[0].duration1 > 0))
          {
            if (rx_items[1].level0 == 0)
            {
                _is_present = true;
            }
          }
        }
      }
      vRingbufferReturnItem(i->rb, (void *)rx_items);
    }
    else
    {
        // time out occurred, this indicates an unconnected / misconfigured bus
        ESP_LOGE(TAG, "rx_items == 0");
        res = OWB_STATUS_HW_ERROR;
    }
  }
  else
  {
      // error in tx channel
      ESP_LOGE(TAG, "Error tx");
      res = OWB_STATUS_HW_ERROR;
  }

  rmt_rx_stop(i->rx_channel);
  rmt_set_rx_idle_thresh(i->rx_channel, old_rx_thresh);

  *is_present = _is_present;

  ESP_LOGD(TAG, "_is_present %d", _is_present);

  return res;
}
//---------------------------------------------------------------------
static owb_status _write_bits(const OneWireBus * bus, uint8_t out, int number_of_bits_to_write)
{
  owb_status status = OWB_STATUS_OK;
  return status;
}
//---------------------------------------------------------------------
static owb_status _read_bits(const OneWireBus * bus, uint8_t *in, int number_of_bits_to_read)
{
  int res = OWB_STATUS_OK;
  return res;
}
//---------------------------------------------------------------------
static struct owb_driver rmt_function_table =
{
  .name = "owb_rmt",
  .uninitialize = _uninitialize,
  .reset = _reset,
  .write_bits = _write_bits,
  .read_bits = _read_bits
};
//---------------------------------------------------------------------
static owb_status _init(owb_rmt_driver_info *info, gpio_num_t gpio_num,
                        rmt_channel_t tx_channel, rmt_channel_t rx_channel)
{
  owb_status status = OWB_STATUS_HW_ERROR;

  info->bus.driver = &rmt_function_table;
  info->tx_channel = tx_channel;
  info->rx_channel = rx_channel;
  info->gpio = gpio_num;

#ifdef OW_DEBUG
  ESP_LOGI(TAG, "RMT TX channel: %d", info->tx_channel);
  ESP_LOGI(TAG, "RMT RX channel: %d", info->rx_channel);
#endif

  rmt_config_t rmt_tx = {0};
  rmt_tx.channel = info->tx_channel;
  rmt_tx.gpio_num = gpio_num;
  rmt_tx.mem_block_num = 1;
  rmt_tx.clk_div = 80;
  rmt_tx.tx_config.loop_en = false;
  rmt_tx.tx_config.carrier_en = false;
  rmt_tx.tx_config.idle_level = 1;
  rmt_tx.tx_config.idle_output_en = true;
  rmt_tx.rmt_mode = RMT_MODE_TX;
  if (rmt_config(&rmt_tx) == ESP_OK)
  {
    rmt_set_source_clk(info->tx_channel, RMT_BASECLK_APB);  // only APB is supported by IDF 4.2
    if (rmt_driver_install(rmt_tx.channel, 0, ESP_INTR_FLAG_LOWMED | ESP_INTR_FLAG_IRAM | ESP_INTR_FLAG_SHARED) == ESP_OK)
    {
      rmt_config_t rmt_rx = {0};
      rmt_rx.channel = info->rx_channel;
      rmt_rx.gpio_num = gpio_num;
      rmt_rx.clk_div = 80;
      rmt_rx.mem_block_num = 1;
      rmt_rx.rmt_mode = RMT_MODE_RX;
      rmt_rx.rx_config.filter_en = true;
      rmt_rx.rx_config.filter_ticks_thresh = 30;
      rmt_rx.rx_config.idle_threshold = OW_DURATION_RX_IDLE;
      if (rmt_config(&rmt_rx) == ESP_OK)
      {
        rmt_set_source_clk(info->rx_channel, RMT_BASECLK_APB);  // only APB is supported by IDF 4.2
        if (rmt_driver_install(rmt_rx.channel, 512, ESP_INTR_FLAG_LOWMED | ESP_INTR_FLAG_IRAM | ESP_INTR_FLAG_SHARED) == ESP_OK)
        {
          rmt_get_ringbuf_handle(info->rx_channel, &info->rb);
          status = OWB_STATUS_OK;
        }
        else
        {
          ESP_LOGE(TAG, "failed to install rx driver");
        }
      }
      else
      {
        status = OWB_STATUS_HW_ERROR;
        ESP_LOGE(TAG, "failed to configure rx, uninstalling rmt driver on tx channel");
        rmt_driver_uninstall(rmt_tx.channel);
      }
    }
    else
    {
      ESP_LOGE(TAG, "failed to install tx driver");
    }
  }
  else
  {
    ESP_LOGE(TAG, "failed to configure tx");
  }

  // attach GPIO to previous pin
  if (gpio_num < 32)
  {
      GPIO.enable_w1ts = (0x1 << gpio_num);
  }
  else
  {
      GPIO.enable1_w1ts.data = (0x1 << (gpio_num - 32));
  }

  rmt_set_gpio(info->rx_channel, RMT_MODE_RX, gpio_num, false);
  rmt_set_gpio(info->tx_channel, RMT_MODE_TX, gpio_num, false);

  // force pin direction to input to enable path to RX channel
  PIN_INPUT_ENABLE(GPIO_PIN_MUX_REG[gpio_num]);

  // enable open drain
  GPIO.pin[gpio_num].pad_driver = 1;


  return status;
}
//---------------------------------------------------------------------
OneWireBus * owb_rmt_initialize(owb_rmt_driver_info * info, gpio_num_t gpio_num,
                                rmt_channel_t tx_channel, rmt_channel_t rx_channel)
{
  ESP_LOGD(TAG, "%s: gpio_num: %d, tx_channel: %d, rx_channel: %d",
             __func__, gpio_num, tx_channel, rx_channel);

  owb_status status = _init(info, gpio_num, tx_channel, rx_channel);
  if (status != OWB_STATUS_OK)
  {
      ESP_LOGE(TAG, "_init() failed with status %d", status);
  }

  info->bus.strong_pullup_gpio = GPIO_NUM_NC;

  return &(info->bus);
}
//---------------------------------------------------------------------
